{- $Id: HaskellMPI.hs,v 1.1.1.2 2001/02/28 04:48:03 michaelw Exp $
 - -*- Haskell -*-
 -----------------------------------------------------------------------------
 - HaskellMPI - (in)complete binding for MPI
 -----------------------------------------------------------------------------
 -
 - hMPI - a Haskell binding for MPI 1.1/1.2
 - Copyright (C) 2000,2001 Michael Weber <michael.weber@post.rwth-aachen.de>
 -
 - This library is free software; you can redistribute it and/or
 - modify it under the terms of the GNU Lesser General Public License
 - as published by the Free Software Foundation; either version 2 of
 - the License, or (at your option) any later version.
 - 
 - This library is distributed in the hope that it will be useful, but
 - WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 - Lesser General Public License for more details.
 - 
 - You should have received a copy of the GNU Lesser General Public
 - License along with this library; if not, write to the Free Software
 - Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 - USA
 - 
 -----------------------------------------------------------------------------
 - NOTES:                                                                      
 - 
 -----------------------------------------------------------------------------
 - TODO:                                                                       
 - * complete the binding
 -
 - * performance check for bcast
 -
 - * check types in interface description
 -
 - * document this file properly                                               
 -
 - * make all MPI ops exception-proof
 - 
 -----------------------------------------------------------------------------
 -}

module HaskellMPI
    ( BRequest       -- abstract
    , send           -- Serialize a => MPI.Rank -> MPI.Tag -> MPI.Comm -> a -> IO ()
    , isend          -- Serialize a => MPI.Rank -> MPI.Tag -> MPI.Comm -> a -> IO (BRequest a)
    , sendrecvMax    -- (Serialize a, Serialize b) 
                     -- => a   -> MPI.Rank -> MPI.Tag 
                     -- -> Int -> MPI.Rank -> MPI.Tag 
                     -- -> MPI.Comm -> IO (MPI.Status,b)
    , recv           -- Serialize a => MPI.Rank -> MPI.Tag -> MPI.Comm -> IO (MPI.Status, a)
    , recvMax        -- Serialize a => Int -> MPI.Rank -> MPI.Tag -> MPI.Comm -> IO (MPI.Status, a)
    , lazySendList   -- Serialize a => MPI.Rank -> MPI.Tag -> MPI.Tag -> MPI.Comm -> [a] -> IO ()
    , lazyRecvList   -- Serialize a => MPI.Rank -> MPI.Tag -> MPI.Comm -> IO [a]
    , killLazySend   -- MPI.Rank -> MPI.Tag -> MPI.Comm -> IO ()
    , irecvMax       -- Serialize a => Int -> MPI.Rank -> MPI.Tag -> MPI.Comm -> IO (BRequest a)
    , bcast          -- Serialize a => MPI.Rank -> MPI.Comm -> a -> IO a
    , bcastMax       -- Serialize a => Int -> MPI.Rank -> MPI.Comm -> a -> IO a
    , wait           -- Serialize a => BRequest a -> IO (MPI.Status,a)
    , wait_          -- BRequest a -> IO MPI.Status
    , test           -- Serialize a => BRequest a -> IO (Maybe (MPI.Status,a))
    , test_          -- BRequest a -> IO (Maybe MPI.Status)
    , waitAll        -- Serialize a => [BRequest a] -> IO [(MPI.Status, a)]
    , waitAll_       -- [BRequest a] -> IO [MPI.Status]
    , probe          -- MPI.Rank -> MPI.Tag -> MPI.Comm -> IO MPI.Status        
    , iprobe         -- MPI.Rank -> MPI.Tag -> MPI.Comm -> IO (Maybe MPI.Status)
    , cancel         -- BRequest a -> IO ()                             
    , reduce         -- Serialize a => MPI.Op -> MPI.Rank -> MPI.Comm -> a -> IO a
    , withHaskellMPI -- (MPI.Comm -> IO a) -> IO a

    , toReq          -- BRequest a -> MPI.Request    
    , module MPI

    , Serialize(..)
    )
where

import qualified MPI
import MPI hiding ( Request(..)
                  , send
                  , isend
                  , sendrecv
                  , recv
                  , irecv
                  , bcast
                  , wait
                  , test
                  , waitAll
                  , reduce
                  , cancel
                  , withMPI
                  )

import MPI_Serialize
import MPI_Types

import Maybe (isJust, fromMaybe)
import IOExts (fixIO, unsafePerformIO)
import Foreign
import Monad (liftM, when)


------------------------------------------------------------------------------
data BRequest a = BRequest { bReqPtr :: MPI.Request
                           , bReqBuf :: Buffer a
                           }

toReq :: BRequest a -> MPI.Request
toReq = bReqPtr


-- HACK! HACK! HACK!
instance Serialize a => Serialize (BRequest a) where
    datatypeOf _ = datatypeOfBRequest undefined
      where
        datatypeOfBRequest :: a -> Datatype
        datatypeOfBRequest = datatypeOf

    sizeInElems = error "sizeInElems: illegal argument of type  (BRequest a)"
    sizeInBytes = error "sizeInBytes: illegal argument of type (BRequest a)"
    serialize   = error "serialize: illegal argument of type (BRequest a)"
    deserialize = error "deserialize: illegal argument of type (BRequest a)"

        
-----------------------------------------------------------------------------
send :: Serialize a => MPI.Rank -> MPI.Tag -> MPI.Comm -> a -> IO ()
send to tag comm what = 
    safeSerialize what
        (\ w -> MPI.send w (sizeInElems what) (datatypeOf what) to tag comm)


isend :: Serialize a => MPI.Rank -> MPI.Tag -> MPI.Comm -> a
      -> IO (BRequest a)
isend to tag comm what = do
    buf <- serialize what
    -- FIXME: free buffer on exception occurrence
    req <- MPI.isend buf (sizeInElems what) (datatypeOf what) to tag comm
    return BRequest{ bReqPtr = req, bReqBuf = buf }


sendrecvMax :: (Serialize a, Serialize b) 
            => a   -> MPI.Rank -> MPI.Tag 
            -> Int -> MPI.Rank -> MPI.Tag 
            -> MPI.Comm -> IO (MPI.Status,b)
sendrecvMax what   dest   stag 
            rcount source rtag comm = sendrecv' undefined
  where
    sendrecv' :: Serialize b => b -> IO (MPI.Status, b)
    sendrecv' unused = do
        safeSerialize what
          (\ sbuf -> do
            let rdt = datatypeOf unused
            rbuf' <- mallocDataBuffer rdt rcount
	    let rbuf = castBuffer rbuf'
            -- FIXME: free buffer on exception occurrence
            stat <- MPI.sendrecv sbuf (sizeInElems what) (datatypeOf what)
                                 dest stag 
                                 rbuf rcount rdt 
                                 source rtag comm

            ret  <- deserialize (fromMaybe 0 (MPI.statusGetCount stat rdt)) 
                                rbuf
            return (stat, ret)
          )
    

recv :: Serialize a => MPI.Rank -> MPI.Tag -> MPI.Comm
     -> IO (MPI.Status, a)
recv from tag comm = recv' undefined
  where
    recv' :: Serialize a => a -> IO (MPI.Status,a)
    recv' unused = do
        let mpiTy = datatypeOf unused
        statp <- probe from tag comm
        let numOfElems = fromMaybe 0 (MPI.statusGetCount statp mpiTy)
	    sFrom      = MPI.statusGetSource statp
	    sTag       = MPI.statusGetTag    statp
        buf   <- mallocDataBuffer mpiTy numOfElems
        -- FIXME: free buffer on exception occurrence
        stat  <- MPI.recv buf numOfElems mpiTy sFrom sTag comm
        ret   <- deserialize (fromMaybe 0 (MPI.statusGetCount stat mpiTy))
                             buf
        return (stat, ret)


recvMax :: Serialize a => Int -> MPI.Rank -> MPI.Tag -> MPI.Comm
        -> IO (MPI.Status, a)
recvMax maxCount from tag comm = recv' undefined
  where
    recv' :: Serialize a => a -> IO (MPI.Status,a)
    recv' unused = do
        let mpiTy = datatypeOf unused
        buf  <- mallocDataBuffer mpiTy maxCount
        -- FIXME: free buffer on exception occurrence
        stat <- MPI.recv buf maxCount mpiTy from tag comm
        ret  <- deserialize (fromMaybe 0 (MPI.statusGetCount stat mpiTy)) buf
        return (stat, ret)


bcast :: Serialize a => MPI.Rank -> MPI.Comm -> a -> IO a
{- 
 - bcast does not work with different lists...
 - Example: Rank 0        : r <- bcastAny ... [1..100]
 -          any other rank: r <- bcastAny ... []
 -          ^-- that one wants to recv the bcast'ed message
 -
 - The above situation would raise a segmentation fault since on Rank r
 - (r > 0) too less memory is allocated!
 -
 - This DIY version of bcast should not suffer from the above
 - mentioned problem.
 -}
dummyTag = MPI.tag 29999
-- `what' may only be evaluated, iff rank == myRank!
bcast rank comm what = do

    -- NOTE: to prevent eavesdropping resp. spoofing we introduce a new
    --       communicator pcomm. [Consider the case of a pending
    --       `irecv/isend ... dummyTag ...'  just before a bcast. it
    --       would most likely interfere...]
    MPI.commDup comm `MPI.withComm` \ pcomm -> do
    
    size   <- MPI.commGetSize pcomm
    myRank <- MPI.commGetRank pcomm

    let receiverList = map MPI.rank [0 .. size-1]

    -- all nodes should have reached this point before we proceed!
    -- (otherwise someone could "steal" transmissions)
    MPI.barrier pcomm -- FIXME: now possibly unnecessary

    -- `rank' denotes the "master", i.e. the one who initiates the
    -- broadcast
    reqs <- if rank == myRank
              then mapM (\ r -> isend r dummyTag pcomm what) receiverList
	      else return []

    (_,recvVal) <- recv rank dummyTag pcomm

    -- clear isend buffers and wait until all isends have finished
    waitAll_ reqs

    return recvVal
 

bcastMax :: Serialize a => Int -> MPI.Rank -> MPI.Comm -> a -> IO a
bcastMax count rank comm what = do
    let mpiTy = datatypeOf what
    myRank <- MPI.commGetRank comm
    buf <- if rank == myRank
             then serialize what
             else mallocDataBuffer mpiTy count
    
    MPI.bcast buf count mpiTy rank comm
    deserialize count buf


-- irecv not very obvious... cannot predict buffer size :-(


irecvMax :: Serialize a => Int -> MPI.Rank -> MPI.Tag -> MPI.Comm 
      -> IO (BRequest a)
irecvMax wantCount from tag comm = irecv' undefined
  where
    irecv' :: Serialize a => a -> IO (BRequest a)
    irecv' unused = do
        let mpiTy = datatypeOf unused
        buf <- mallocDataBuffer mpiTy wantCount
        -- FIXME: free buffer on exception occurrence
        req <- MPI.irecv buf wantCount mpiTy from tag comm
        return BRequest{ bReqPtr = req, bReqBuf = buf }


wait :: Serialize a => BRequest a -> IO (MPI.Status,a)
-- NOTE:

-- it is not legal to apply `wait' to a BRequest, which resulted from
-- a MPI_PROC_NULL-call, since it tries to deserialize something and a
-- (count == 0) is not allowed in deserialize (except for list
-- versions).
--
-- other possible solutions:
-- * use Maybe return type:
--
--   not very convenient, one can almost achieve the same by trying to
--   receive a singleton list (by ``(s,v::[a]) <- irecvMax 1 ...'').
--   If it's empty, then the call use NullProcess as destination.
--
-- * throw an exception:
--
--   not very application programmer friendly, too. In almost every
--   situation, one would have to shield the wait with a `catch'
--   statement, since this is not real error.
--
-- For mutable objects the case is different, here a call to
-- `deserialize' should not alter the object. So the original value is
-- preserved.
-- 
-- To conclude: this type of programming scheme is not very
-- compatible with Haskell =) Try to avoid it, or know what you're
-- doing...
--
wait req = do
    stat  <- MPI.wait (bReqPtr req)
    let mpiTy = datatypeOf req
        count = fromMaybe 0 (MPI.statusGetCount stat mpiTy)
    val <- deserialize count (bReqBuf req)
    return (stat, val)


wait_ :: BRequest a -> IO MPI.Status
wait_ req = do
    stat <- MPI.wait (bReqPtr req)
    freeBuffer (bReqBuf req)
    return stat


test :: Serialize a => BRequest a -> IO (Maybe (MPI.Status,a))
test req = do
    mbStat <- MPI.test (bReqPtr req)
    case mbStat of
      Nothing   -> return Nothing
      Just stat -> do
          val   <- deserialize (fromMaybe 0 
                                  (MPI.statusGetCount stat (datatypeOf req)))
                               (bReqBuf req)
          return (Just (stat, val))


test_ :: BRequest a -> IO (Maybe MPI.Status)
test_ req = do
    mbStat <- MPI.test (bReqPtr req)
    when (isJust mbStat) (freeBuffer (bReqBuf req))
    return mbStat


waitAll :: Serialize a => [BRequest a] -> IO [(MPI.Status, a)]
waitAll = mapM wait -- FIXME


waitAll_ :: [BRequest a] -> IO [MPI.Status]
waitAll_ = mapM wait_ -- FIXME
           

-- probe  = MPI.probe
-- iprobe = MPI.iprobe


cancel req = do
    MPI.cancel (bReqPtr req)
    freeBuffer (bReqBuf req)


reduce :: Serialize a => MPI.Op -> MPI.Rank -> MPI.Comm -> a -> IO a
-- FIXME: use MPI_IN_PLACE where avail.
reduce op rank comm what = do
    safeSerialize 
        what
        (\ w -> do
            w' <- mallocByteBuffer (sizeInBytes what)
            MPI.reduce w w' (sizeInElems what) (datatypeOf what) op rank comm
            deserialize (sizeInElems what) w'
        )


withHaskellMPI :: (MPI.Comm -> IO a) -> IO a
withHaskellMPI p = MPI.withMPI (p MPI.CommWorld)


-----------------------------------------------------------------------------
dataDontCare = MPI.DataInt

lazySendList :: Serialize a 
             => MPI.Rank -> MPI.Tag 
             -> MPI.Rank -> MPI.Tag -> MPI.Comm -> [a] -> IO ()
lazySendList from killTag 
             to   tag comm what = do
    req <- MPI.irecv (toBuffer nullPtr) 0 dataDontCare from killTag comm
    doSend req what
  where
    doSend req []     = MPI.cancel req
    doSend req (v:vs) = do
      breq <- isend to tag comm v
      (status, rest) <- MPI.waitAny [req,toReq breq]
      if statusGetTag status == killTag
        then cancel breq
        else doSend req vs


lazyRecvList :: Serialize a => MPI.Rank -> MPI.Tag -> MPI.Comm -> IO [a]
lazyRecvList from tag comm = recv'
  where
    recv' :: Serialize a => IO [a]
    recv' = do
        (_,v) <- recv from tag comm
        return (v : unsafePerformIO recv')


killLazySend :: MPI.Rank -> MPI.Tag -> MPI.Comm -> IO ()
killLazySend = MPI.send (toBuffer nullPtr) 0 dataDontCare


-- helper stuff, not for export ----------------------------------------------
------------------------------------------------------------------------------
