{- $Id: MPI_Serialize.hs,v 1.1.1.2 2001/02/28 04:48:03 michaelw Exp $
 - -*- Haskell -*-
 -----------------------------------------------------------------------------
 - MPI_Serialize.hs - data marshalling for the MPI binding
 -----------------------------------------------------------------------------
 -
 - hMPI - a Haskell binding for MPI 1.1/1.2
 - Copyright (C) 2000,2001 Michael Weber <michael.weber@post.rwth-aachen.de>
 -
 - This library is free software; you can redistribute it and/or
 - modify it under the terms of the GNU Lesser General Public License
 - as published by the Free Software Foundation; either version 2 of
 - the License, or (at your option) any later version.
 - 
 - This library is distributed in the hope that it will be useful, but
 - WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 - Lesser General Public License for more details.
 - 
 - You should have received a copy of the GNU Lesser General Public
 - License along with this library; if not, write to the Free Software
 - Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 - USA
 - 
 -----------------------------------------------------------------------------
 - NOTES:                                                                      
 -
 -----------------------------------------------------------------------------
 - TODO:                                                                       
 -
 -----------------------------------------------------------------------------
 -}

module MPI_Serialize
    ( module MPI_Serialize
    )
where

import MPI_Types 
    ( Datatype
    , Request )
import MPI_MarshalInternal
import MPI_Datatypes
    ( typeExtent )
import MPI_Consts 
    ( hmpi_bottom )

import Monad
import Exception
    ( bracket )
import Foreign
import Foreign.Storable

-----------------------------------------------------------------------------
type Buffer a = MPI_Buffer a

toBuffer :: Ptr a -> Buffer a
toBuffer = MPI_Buffer


bufferToAddr :: Buffer a -> Ptr a
bufferToAddr (MPI_Buffer ptr) = ptr

mallocElem  :: Storable a => a           -> IO (Ptr a)
mallocElems :: Storable a => a    -> Int -> IO (Ptr a)
mallocElem      = mallocBytes . sizeOf
mallocElems x n = mallocBytes (n * sizeOf x)

freeBuffer :: Buffer a -> IO ()
freeBuffer (MPI_Buffer ptr) = free ptr


nullBuffer :: Buffer a
nullBuffer = toBuffer nullPtr


bottomBuffer :: Buffer a
bottomBuffer = hmpi_bottom


mallocByteBuffer :: Int -> IO (Buffer a)
mallocByteBuffer count = do
    liftM MPI_Buffer (mallocBytes count)


mallocDataBuffer :: Datatype -> Int -> IO (Buffer a)
mallocDataBuffer mpiTy count = do
    tySizeInBytes <- typeExtent mpiTy
    liftM MPI_Buffer (mallocBytes (count * (fromIntegral tySizeInBytes)))


------------------------------------------------------------------------------
class Serialize a where
    datatypeOf     :: a -> Datatype
    sizeInElems    :: a -> Int -- size in MPI units

    sizeInBytes    :: a -> Int
    serialize      :: a -> IO (Buffer a)
    deserialize    :: Int -> (Buffer a) -> IO a

    -- This should be a common case, except for lists of elements...
    sizeInElems            = const 1


-----------------------------------------------------------------------------
safeSerialize :: Serialize a => a -> (Buffer a -> IO b) -> IO b
safeSerialize what cmd = do
    bracket (serialize what)
            freeBuffer
            cmd


-----------------------------------------------------------------------------
serializeElem :: Storable a => a -> IO (Buffer a)
serializeElem x = do
    ptr <- mallocElem x
    poke ptr x
    return (MPI_Buffer (castPtr ptr))


deserializeElem :: Storable a => Buffer a -> IO a
deserializeElem (MPI_Buffer ptr) = do
   v <- peek ptr
   free ptr
   return v


serializeList :: Storable a => [a] -> IO (Buffer a)
serializeList ls = do
    ptr <- mallocElems (head ls) (length ls)
    mapM_ (uncurry (pokeElemOff ptr)) (zip [0 ..] ls)
    return (MPI_Buffer ptr)


deserializeList :: Storable a => Int -> (Buffer a) -> IO [a]
deserializeList len (MPI_Buffer ptr) = do
  v <- mapM (peekElemOff ptr) [0 .. len - 1]
  free ptr
  return v


-----------------------------------------------------------------------------
newtype ZString = ZString [Char]
    deriving (Eq, Read, Show)


toZString                 = ZString
fromZString (ZString str) = str


serializeZString :: ZString -> IO (Buffer a)
serializeZString (ZString str) = do
    let len = length str
    ptr <- mallocElems (head str) (len+1)
    mapM_ (uncurry (pokeElemOff ptr)) (zip [0 ..] str)
    pokeElemOff ptr len '\0'
    return (MPI_Buffer (castPtr ptr))


deserializeZString :: (Buffer a) -> IO ZString
deserializeZString (MPI_Buffer ptr) = liftM ZString (dZStr 0)
  where  
    dZStr ofs = do
       c <- peekElemOff (castPtr ptr) ofs
       if c == '\0' 
         then do
           free ptr
           return ""
         else liftM (c :) (dZStr (ofs + 1))


serializeLString :: String -> IO (Buffer a)
serializeLString str = do
    let len = length str
    ptr <- mallocElems (head str) len
    mapM_ (uncurry (pokeElemOff ptr)) (zip [0 .. len - 1] str)
    return (MPI_Buffer (castPtr ptr))


deserializeLString :: Int -> (Buffer a) -> IO String
deserializeLString len (MPI_Buffer ptr) = do
  v <- unmarshalStringWithLen (castPtr ptr) len
  free ptr
  return v


------------------------------------------------------------------------------
