{- $Id: MPI_Transfer.hs,v 1.1.1.2 2001/02/28 04:48:03 michaelw Exp $
 - -*- Haskell -*-
 -----------------------------------------------------------------------------
 - MPI_Transfer.hs - sending/receiving malloc'ed data via MPI
 -----------------------------------------------------------------------------
 -
 - hMPI - a Haskell binding for MPI 1.1/1.2
 - Copyright (C) 2000,2001 Michael Weber <michael.weber@post.rwth-aachen.de>
 -
 - This library is free software; you can redistribute it and/or
 - modify it under the terms of the GNU Lesser General Public License
 - as published by the Free Software Foundation; either version 2 of
 - the License, or (at your option) any later version.
 - 
 - This library is distributed in the hope that it will be useful, but
 - WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 - Lesser General Public License for more details.
 - 
 - You should have received a copy of the GNU Lesser General Public
 - License along with this library; if not, write to the Free Software
 - Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 - USA
 - 
 -----------------------------------------------------------------------------
 - NOTES:                                                                      
 - 
 -----------------------------------------------------------------------------
 - TODO:                                                                       
 - * document this file properly                                               
 -----------------------------------------------------------------------------
 -}

module MPI_Transfer
    ( send
    , recv
    , sendrecv
    , bcast
    , isend
    , irecv
    , wait
    , test
    , probe
    , iprobe
    , cancel
    , testCancelled
    , requestFree
    , requestGetStatus
    , waitAny
    , testAny
    , waitAll
    , testAll
    , testSome
--     , waitSome
    )
where

import MPI_Types
import MPI_CCalls
import MPI_MarshalInternal

import Monad
import Maybe


------------------------------------------------------------------------------
send :: MPI_Buffer a -> Int -> Datatype -> Rank -> Tag -> Comm -> IO ()
send buf count dt rank tag comm = do
    safeCall_ (mpi_Send buf 
                   (fromIntegral count) 
                   (marshalDatatype dt) 
                   (marshalRank rank)
                   (marshalTag tag)
                   (marshalComm comm))


recv :: MPI_Buffer a -> Int -> Datatype -> Rank -> Tag -> Comm -> IO Status
recv buf count dt rank tag comm = do
    status <- newStatus
    safeCall_ (mpi_Recv buf 
                   (fromIntegral count) 
                   (marshalDatatype dt) 
                   (marshalRank rank)
                   (marshalTag tag)
                   (marshalComm comm)
                   (statusPtr status))
    return status


sendrecv :: MPI_Buffer a -> Int -> Datatype -> Rank -> Tag 
         -> MPI_Buffer a -> Int -> Datatype -> Rank -> Tag 
         -> Comm -> IO Status
sendrecv sbuf scount sdt dest   stag 
         rbuf rcount rdt source rtag comm = do
    status <- newStatus
    safeCall_ (mpi_Sendrecv sbuf 
                   (fromIntegral scount) 
                   (marshalDatatype sdt) 
                   (marshalRank dest)
                   (marshalTag stag)
                   rbuf
                   (fromIntegral rcount) 
                   (marshalDatatype rdt) 
                   (marshalRank source)
                   (marshalTag rtag)
                   (marshalComm comm)
                   (statusPtr status))
    return status


bcast :: MPI_Buffer a -> Int -> Datatype -> Rank -> Comm -> IO ()
bcast buf count dt rank comm = do
    safeCall_ (mpi_Bcast buf 
                   (fromIntegral count) 
                   (marshalDatatype dt) 
                   (marshalRank rank)
                   (marshalComm comm))


isend :: MPI_Buffer a -> Int -> Datatype -> Rank -> Tag -> Comm -> IO Request
isend buf count dt rank tag comm = do
    req <- newRequest
    safeCall_ (mpi_Isend buf 
                   (fromIntegral count) 
                   (marshalDatatype dt) 
                   (marshalRank rank)
                   (marshalTag tag)
                   (marshalComm comm)
                   (reqPtr req))
    return req


irecv :: MPI_Buffer a -> Int -> Datatype -> Rank -> Tag -> Comm -> IO Request
irecv buf count dt rank tag comm = do
    req <- newRequest
    safeCall_ (mpi_Irecv buf 
                   (fromIntegral count) 
                   (marshalDatatype dt) 
                   (marshalRank rank)
                   (marshalTag tag)
                   (marshalComm comm)
                   (reqPtr req))
    return req


wait :: Request -> IO Status
wait req = do
    status <- newStatus
    safeCall_ (mpi_Wait (reqPtr req) (statusPtr status))
    clearRequestFinalizer req
    return status


test :: Request -> IO (Maybe Status)
test req = do
    status <- newStatus
    flag   <- liftM unmarshalBool (do
        tempAllocOut_ (\ pflag -> do
        safeCall_ (mpi_Test (reqPtr req) pflag (statusPtr status))
        ))
    unless flag (clearRequestFinalizer req)
    return (if flag then Just status else Nothing)


probe :: Rank -> Tag -> Comm -> IO Status
probe rank tag comm = do
    status <- newStatus
    safeCall_ (mpi_Probe (marshalRank rank) 
                         (marshalTag tag)
                         (marshalComm comm) 
                         (statusPtr status))
    return status


iprobe :: Rank -> Tag -> Comm -> IO (Maybe Status)
iprobe rank tag comm = do
    status <- newStatus
    flag   <- liftM unmarshalBool (do
        tempAllocOut_ (\ pflag -> do
        safeCall_ (mpi_Iprobe (marshalRank rank) 
                              (marshalTag tag)
                              (marshalComm comm) 
                              pflag
                              (statusPtr status))
        ))
    return (if flag then Just status else Nothing)


cancel :: Request -> IO ()
cancel req = 
    safeCall_ (mpi_Cancel (reqPtr req))


testCancelled :: Status -> IO Bool
testCancelled status = liftM unmarshalBool (do
    tempAllocOut_ (safeCall_ . (mpi_Test_cancelled (statusPtr status))))


requestFree :: Request -> IO ()
requestFree req = do
    safeCall_ (mpi_Request_free (reqPtr req))
    clearRequestFinalizer req


requestGetStatus :: Request -> IO (Maybe Status)
requestGetStatus req = do
    status <- newStatus
    flag   <- liftM unmarshalBool (do
        tempAllocOut_ (\ pflag -> do
        safeCall_ (mpi_Request_get_status (reqPtr req) 
                                          pflag
                                          (statusPtr status))
        ))
    return (if flag then Just status else Nothing)
     

{- Unfortunately, we cannot easily use MPI's {wait,test}* routines,
 - since they require the requests to be in a straight array of type
 - `MPI_Request* array_of_requests'...
 - Remember: most of these functions _modify_ the request array.
 - 
 - Therefore we provide native Haskell functions, which mimic the
 - original behaviour.
 -}
waitAny :: [Request] -> IO (Status, [Request])
waitAny []       = ioError $ userError "MPI.waitAny: empty list"
waitAny rs@(_:_) = do
    r <- testAny rs
    case r of
      Nothing -> waitAny rs
      Just ok -> return ok


testAny :: [Request] -> IO (Maybe (Status, [Request]))
testAny rs = testAny' rs []
  where
    testAny' []     _  = return Nothing
    testAny' (r:rs) ps = do
        ret <- test r
        case ret of
          Nothing -> testAny' rs (r:ps)
          Just x  -> return $ Just (x, reverse ps ++ rs)


waitAll :: [Request] -> IO [Status]
waitAll = mapM wait -- FIXME


testAll :: [Request] -> IO [Status]
testAll rs = do -- FIXME
#ifdef HAVE_MPI_REQUEST_GET_STATUS

    rs' <- mapM requestGetStatus rs

#else /* HAVE_MPI_REQUEST_GET_STATUS */

    {- This is a change in semantics, but IMHO cannot be avoided until 
     - `requestGetStatus' is available.
     -
     - `testAll' may not change the status of *any* Request object, unless
     - all of them indicate a finished communication. However, `test'
     - frees a "ready" Request immediately.
     -} 
    rs' <- mapM test rs

#endif /* HAVE_MPI_REQUEST_GET_STATUS */

    if any (== Nothing) rs'
      then return []
      else do -- mapM_ requestFree rs -- FIXME: not needed?
              return $ map fromJust rs'


testSome :: [Request] -> IO [(Status,Request)]
testSome rs = do
    rs' <- mapM test rs
    return $ mapMaybe sync (zip rs' rs)
  where
    sync (Nothing, y) = Nothing
    sync (Just x,  y) = Just (x,y)


-- waitSome :: [Request] -> IO [(Status,Request)]
-- waitSome []       = ioError $ userError "MPI.waitSome: empty list"
-- waitSome rs@(_:_) = do
--     r <- testSome rs
--     case r of
--       [] -> waitSome rs
--       ok -> return ok


------------------------------------------------------------------------------
