{- $Id: MPI_Types.hs,v 1.1.1.2 2001/02/28 04:48:03 michaelw Exp $
 - -*- Haskell -*-
 -----------------------------------------------------------------------------
 - MPI_Types.hs - MPI types and their respective marshalling functions
 -----------------------------------------------------------------------------
 -
 - hMPI - a Haskell binding for MPI 1.1/1.2
 - Copyright (C) 2000,2001 Michael Weber <michael.weber@post.rwth-aachen.de>
 -
 - This library is free software; you can redistribute it and/or
 - modify it under the terms of the GNU Lesser General Public License
 - as published by the Free Software Foundation; either version 2 of
 - the License, or (at your option) any later version.
 - 
 - This library is distributed in the hope that it will be useful, but
 - WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 - Lesser General Public License for more details.
 - 
 - You should have received a copy of the GNU Lesser General Public
 - License along with this library; if not, write to the Free Software
 - Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 - USA
 - 
 -----------------------------------------------------------------------------
 - NOTES:                                                                      
 - * this files contains the basic MPI metadata types.  See MPI docs for
 -   a description what they are good for.
 -
 - * since 0.9.0.1: 
 -     changed Rank/Tag so that MPI_Rank/MPI_Tag is stripped.
 -     This makes it easier to pattern match over them.
 -
 -----------------------------------------------------------------------------
 - TODO:                                                                       
 - * because of a possible loop in the module structure of hMPI this
 -   module also contains some stuff that really belongs to
 -   MPI_MarshalInternal.
 -
 - * document this file properly                                               
 -----------------------------------------------------------------------------
 -}

module MPI_Types
    ( Rank(..)
    , rank
    , marshalRank
    , unmarshalRank
     
    , Tag(..) 
    , tag
    , marshalTag
    , unmarshalTag
    , allocaElem, allocaElems, pokeElemOff, peekElemOff
    , castBuffer

    , Comm
          ( CommNull
          , CommSelf
          , CommWorld
          )
    , marshalComm
    , unmarshalComm

    , Group(GroupEmpty)
    , marshalGroup
    , unmarshalGroup

    , Compare(..)
    , unmarshalCompare

    , Status, statusPtr
    , newStatus

    , Request
    , newRequest
    , reqPtr
    , setRequestFinalizer
    , clearRequestFinalizer
    , RequestException(StrayRequest)

    , Error, errorVal
    , newError

    , Datatype
          ( DataNull
          , DataChar
          , DataUnsignedChar
          , DataByte
          , DataShort
          , DataUnsignedShort
          , DataInt
          , DataUnsigned
          , DataLong
          , DataUnsignedLong
          , DataFloat
          , DataDouble
          , DataLongDouble
          , DataLongLongInt
          , DataPacked
          , DataLB
          , DataUB
          , DataFloatInt
          , DataDoubleInt
          , DataLongInt
          , DataShortInt
          , Data2Int
          , DataLongDoubleInt
          )
    , marshalDatatype
    , unmarshalDatatype

    , Op(..)
    , marshalOp

    , Version, versionMajor, versionMinor
    , newVersion

    , SystemError
          ( SystemSuccess      
          , SystemErrorBuffer  
          , SystemErrorCount   
          , SystemErrorType    
          , SystemErrorTag     
          , SystemErrorComm    
          , SystemErrorRank    
          , SystemErrorRoot    
          , SystemErrorGroup   
          , SystemErrorOp      
          , SystemErrorTopology
          , SystemErrorDims    
          , SystemErrorArg     
          , SystemErrorUnknown 
          , SystemErrorTruncate
          , SystemErrorOther   
          , SystemErrorIntern  
          , SystemErrorInStatus
          , SystemErrorPending 
          , SystemErrorRequest 
          , SystemError
          )

    , tempAllocOut
    , tempAllocOut_
    , tempAllocInout_
    , tempAllocIn
    , safeCall
    , safeCall_

    , mpiUndefined
    )
where

import MPI_Consts
import MPI_MarshalInternal

import IO hiding (bracket)
import IOExts
import Dynamic
import Exception
import Monad
import Foreign
import Foreign.C
import Foreign.Ptr
#include <config.h>


castBuffer :: MPI_Buffer a -> MPI_Buffer b
castBuffer (MPI_Buffer a) = MPI_Buffer (castPtr a)

-- Rank ----------------------------------------------------------------------
data Rank = AnySource 
          | NullProcess 
          | Rank MPI_Rank_Type
    deriving (Eq, Show, Read)

instance Enum Rank where
    fromEnum = fromEnum . marshalRank
    toEnum   = unmarshalRank . toEnum

    pred AnySource   = NullProcess
    pred NullProcess = NullProcess
    pred (Rank 0)    = NullProcess
    pred (Rank x)    = Rank (pred x)
                     
    succ AnySource   = NullProcess
    succ NullProcess = NullProcess -- not sure... Rank (MPI_Rank 0)?
    succ (Rank x)    = Rank (succ x)


rank :: Integral a => a  -> Rank
rank = Rank . fromIntegral


marshalRank :: Rank -> MPI_Rank
marshalRank AnySource   = hmpi_any_source
marshalRank NullProcess = hmpi_proc_null
marshalRank (Rank r)    = MPI_Rank r


unmarshalRank :: MPI_Rank -> Rank
unmarshalRank r@(MPI_Rank r')
  | r == hmpi_any_source = AnySource
  | r == hmpi_proc_null  = NullProcess
  | otherwise            = Rank r'


-- Tag -----------------------------------------------------------------------
data Tag = AnyTag 
         | Tag MPI_Tag_Type
    deriving (Eq, Show, Read)

instance Enum Tag where
    fromEnum = fromEnum . marshalTag
    toEnum   = unmarshalTag . toEnum


tag :: Integral a => a  -> Tag
tag = Tag . fromIntegral


marshalTag :: Tag -> MPI_Tag
marshalTag AnyTag  = hmpi_any_tag
marshalTag (Tag t) = MPI_Tag t


unmarshalTag :: MPI_Tag -> Tag
unmarshalTag t@(MPI_Tag t')
  | t == hmpi_any_tag = AnyTag
  | otherwise         = Tag t'


-- Comm ----------------------------------------------------------------------
data Comm = CommNull 
          | CommSelf 
          | CommWorld 
          | Comm MPI_Comm


marshalComm CommNull  = hmpi_comm_null
marshalComm CommSelf  = hmpi_comm_self
marshalComm CommWorld = hmpi_comm_world
marshalComm (Comm x)  = x


unmarshalComm x = Comm x


-- Group ---------------------------------------------------------------------
data Group = GroupEmpty
           | Group MPI_Group

marshalGroup GroupEmpty = hmpi_group_empty
marshalGroup (Group x)  = x

unmarshalGroup x = Group x


-- Status --------------------------------------------------------------------
newtype Status = Status{ statusPtr :: MPI_Status } deriving (Eq)

newStatus :: IO Status
-- Description:
--   Creates a new Status Object. The memory under its control gets
--   automatically freed, when the status object is garbage-collected.
newStatus = do
    ptr <- mallocBytes hmpi_sizeof_status
    fo  <- newForeignPtr ptr (free ptr)
    return Status{ statusPtr = MPI_Status fo }


-- Request -------------------------------------------------------------------
data Request =
    Request{ reqPtr       :: MPI_Request
           , reqFinalizer :: IORef RequestFinalizer
           }

instance Eq Request where
    r1 == r2 = reqPtr r1 == reqPtr r2


type RequestFinalizer = MPI_Request -> IO ()


data RequestException = StrayRequest MPI_Request

requestExceptionTc = mkTyCon "RequestException"

instance Typeable RequestException where
  typeOf _ = mkAppTy requestExceptionTc []



newRequest :: IO Request
newRequest = newFRequest requestFinalizeThrowException


newFRequest :: RequestFinalizer -> IO Request
-- Description:
--   Some magic to make automatic finalization of Requests work.  The
--   reason is, that you can partly free the Request somewhere in the
--   Program (this happens when calling wait/test on it). After that,
--   the default finalizer is cleared.  If wait/test is not called on
--   a Request until all pointers to it are removed (a stray request),
--   an error is thrown.
newFRequest finalizer = fixIO $ \ fReq -> do
    ptr <- mallocBytes hmpi_sizeof_request
    fo  <- newForeignPtr ptr (reqFree fReq ptr)
    f   <- newIORef finalizer
    return Request{ reqPtr       = MPI_Request fo
                  , reqFinalizer = f
                  }
  where
    reqFree req ptr = do
        f <- readIORef (reqFinalizer req)
        f (reqPtr req)
        free ptr


requestFinalizeThrowException :: RequestFinalizer
requestFinalizeThrowException ptr = do
#ifdef HMPI_DEBUG
    putStrLn "STRAY REQUEST!" -- FIXME: remove/make configurable!
#endif    
    throwDyn (StrayRequest ptr)


requestFinalizeNone :: RequestFinalizer
requestFinalizeNone _ = return ()


setRequestFinalizer :: Request -> RequestFinalizer -> IO ()
setRequestFinalizer req = writeIORef (reqFinalizer req)


clearRequestFinalizer :: Request -> IO ()
clearRequestFinalizer req = setRequestFinalizer req requestFinalizeNone


-- Error ---------------------------------------------------------------------
newtype Error = Error{ errorVal :: MPI_Error }

newError err = Error{ errorVal = err }


-- Compare -------------------------------------------------------------------
data Compare = Ident
             | Congruent
             | Similar
             | Unequal
    deriving (Eq)


unmarshalCompare :: CInt -> Compare
unmarshalCompare x
  | x == hmpi_compare_ident     = Ident
  | x == hmpi_compare_congruent = Congruent
  | x == hmpi_compare_similar   = Similar
  | x == hmpi_compare_unequal   = Unequal
  | otherwise                   
  = error "MPI_Types.unmarshalCompare: unknown value"


-- Types ---------------------------------------------------------------------
data Datatype = DataNull
              | DataChar
              | DataUnsignedChar
              | DataByte
              | DataShort
              | DataUnsignedShort
              | DataInt
              | DataUnsigned
              | DataLong
              | DataUnsignedLong
              | DataFloat
              | DataDouble        
              | DataLongDouble
              | DataLongLongInt
              | DataPacked
              | DataLB
              | DataUB
              | DataFloatInt
              | DataDoubleInt
              | DataLongInt
              | DataShortInt
              | Data2Int
              | DataLongDoubleInt
              | Datatype MPI_Datatype
    deriving (Show)

marshalDatatype :: Datatype -> MPI_Datatype
marshalDatatype DataNull          = hmpi_datatype_null
marshalDatatype DataChar          = hmpi_datatype_char
marshalDatatype DataUnsignedChar  = hmpi_datatype_uchar
marshalDatatype DataByte          = hmpi_datatype_byte
marshalDatatype DataShort         = hmpi_datatype_short
marshalDatatype DataUnsignedShort = hmpi_datatype_ushort
marshalDatatype DataInt           = hmpi_datatype_int
marshalDatatype DataUnsigned      = hmpi_datatype_unsigned
marshalDatatype DataLong          = hmpi_datatype_long
marshalDatatype DataUnsignedLong  = hmpi_datatype_ulong
marshalDatatype DataFloat         = hmpi_datatype_float
marshalDatatype DataDouble        = hmpi_datatype_double
marshalDatatype (Datatype d)      = d


unmarshalDatatype :: MPI_Datatype -> Datatype
unmarshalDatatype x = Datatype x


-- Op ------------------------------------------------------------------------
data Op = OpMax
        | OpMin
        | OpSum
        | OpProd
        | OpLAND
        | OpBAND
        | OpLOR
        | OpBOR
        | OpLXOR
        | OpBXOR
        | OpMinLoc
        | OpMaxLoc


marshalOp :: Op -> MPI_Op
marshalOp OpMax    = hmpi_op_max
marshalOp OpMin    = hmpi_op_min
marshalOp OpSum    = hmpi_op_sum
marshalOp OpProd   = hmpi_op_prod
marshalOp OpLAND   = hmpi_op_land
marshalOp OpBAND   = hmpi_op_band
marshalOp OpLOR    = hmpi_op_lor
marshalOp OpBOR    = hmpi_op_bor
marshalOp OpLXOR   = hmpi_op_lxor
marshalOp OpBXOR   = hmpi_op_bxor
marshalOp OpMinLoc = hmpi_op_minloc
marshalOp OpMaxLoc = hmpi_op_maxloc


-- Version -------------------------------------------------------------------
data Version = Version { versionMajor :: Int
		       , versionMinor :: Int
		       }

instance Show Version where
    showsPrec _ (Version major minor) =
        showsPrec 0 major .
	showString "."    .
	showsPrec 0 minor

newVersion x y = Version { versionMajor = x
                         , versionMinor = y
                         }


-- Errorcodes returned by MPI functions --------------------------------------
-- NB: nobody has to deal with them at the moment, but they may get
--     back under user control via exceptions... later!

data SystemError = SystemSuccess
                 | SystemErrorBuffer
                 | SystemErrorCount
                 | SystemErrorType
                 | SystemErrorTag
                 | SystemErrorComm
                 | SystemErrorRank
                 | SystemErrorRoot
                 | SystemErrorGroup
                 | SystemErrorOp
                 | SystemErrorTopology
                 | SystemErrorDims
                 | SystemErrorArg
                 | SystemErrorUnknown
                 | SystemErrorTruncate
                 | SystemErrorOther
                 | SystemErrorIntern
                 | SystemErrorInStatus
                 | SystemErrorPending
                 | SystemErrorRequest
                 | SystemError MPI_ReturnCode  -- intentionally here
                 | SystemErrorLastCode 
    deriving (Eq,Show)


systemErrorTc = mkTyCon "SystemError"

instance Typeable SystemError where
  typeOf _ = mkAppTy systemErrorTc []


unmarshalSystemError :: MPI_ReturnCode -> SystemError
unmarshalSystemError x
    | x == hmpi_system_success = SystemSuccess
    | otherwise                = SystemError x


-----------------------------------------------------------------------------
mpiUndefined :: Integral a => a
mpiUndefined = fromIntegral hmpi_undefined

allocaElem  x   op = allocaBytes (sizeOf x)     op
allocaElems x n op = allocaBytes (n * sizeOf x) op
{-
pokeElemOff addr idx x = 
    let off = fromIntegral (idx * sizeOf x)
    in  poke (addr `plusPtr` off) x
peekElemOff addr idx = IOExts.fixIO $ \result ->
		       let off = fromIntegral (idx * sizeOf result)
		       in  peek (addr `plusPtr` off)
-}
-----------------------------------------------------------------------------
-- Unfortunately, these function must be here, although they really
-- belong to the MPI_MarshalInternal-Layer.  However, They use
-- ``SystemError'' and moving them would lead to a circular
-- dependency. :(

tempAllocOut :: Storable a 
             => (Ptr a -> IO r)
             -> IO (a,r)
-- Description: 
--   Calls a function with one out-parameter, reserving temporary
--   space for it.
-- Result: 
--   returncode of the function and the unmarshalled out-Parameter
tempAllocOut mpiCall = fixIO (\ ~(fV,_) -> do
    allocaElem fV (\ ptr -> do
    r <- mpiCall ptr
    v <- peek ptr
    return (v,r)))


tempAllocOut_ :: Storable a 
             => (Ptr a -> IO ()) 
             -> IO a
-- Description: 
--   Calls a function with one out-parameter, reserving temporary
--   space for it.  The returncode of the function is thrown away.
-- Result: 
--   the unmarshalled out-Parameter
tempAllocOut_ mpiCall = fixIO (\ fV -> do
    allocaElem fV (\ ptr -> do
    mpiCall ptr
    peek ptr))


tempAllocInout_ :: Storable a 
             => a 
             -> (Ptr a -> IO ()) 
             -> IO a
-- Description:
--   Calls a function with one inout-parameter (call-by-reference),
--   reserving temporary memory for it.
-- Result:
--   the new value of the parameter
tempAllocInout_ val mpiCall = do
    allocaElem val (\ ptr -> do
    poke ptr val
    mpiCall ptr
    peek ptr)


tempAllocIn :: Storable a 
            => a 
            -> (Ptr a -> IO r) 
            -> IO r
-- Description:
--   Calls a function which accepts a /pointer/ to a value as
--   parameter.
-- Result:
--   the returncode of the function
tempAllocIn val mpiCall = do
    allocaElem val (\ ptr -> do
    poke ptr val
    mpiCall ptr)


-----------------------------------------------------------------------------
safeCall :: (a -> b) -> IO (MPI_ReturnCode,a) -> IO b
-- Description:
--   Calls a function and converts the returncode into an exception.
--   An additional value is piped through a conversion function, which
--   is usually some kind of simple unmarshalling.
-- Restrictions:
--   This is tailor-made for MPI. Don't use it for other functions.
-- Result:
--   Returns the conversion function applied to the value of the
--   called function.
safeCall f mpiCall = do
    (r,v) <- mpiCall
    unless (unmarshalSystemError r == SystemSuccess) (throwDyn r)
    return (f v)


safeCall_ :: IO MPI_ReturnCode -> IO ()
-- Description:
--   calls a function and converts the returncode into an exception.
-- Restrictions:
--   This is tailor-made for MPI. Don't use it for other functions.
-- Result:
--   ()
safeCall_ mpiCall = do
    r <- mpiCall
    unless (unmarshalSystemError r == SystemSuccess) (throwDyn r)


------------------------------------------------------------------------------
