{- $Id: TermCheckMPI.hs,v 1.1.1.1 2001/02/28 04:48:04 michaelw Exp $
 - -*- Haskell -*-
 ----------------------------------------------------------------------------
 - TermCheckMPI.hs - distributed termination checking
 ----------------------------------------------------------------------------
 - Copyright (C) 2000  Michael Weber <michael.weber@post.rwth-aachen.de>
 - 
 - This library is free software; you can redistribute it and/or
 - modify it under the terms of the GNU Lesser General Public License
 - as published by the Free Software Foundation; either version 2 of
 - the License, or (at your option) any later version.
 - 
 - This library is distributed in the hope that it will be useful, but
 - WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 - Lesser General Public License for more details.
 - 
 - You should have received a copy of the GNU Lesser General Public
 - License along with this library; if not, write to the Free Software
 - Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 - USA
 - 
 ----------------------------------------------------------------------------
 - Created: 2000-09-25
 ----------------------------------------------------------------------------
 - NOTES:
 - * modelled after the DFG token termination algorithm
 -   [Dijkstra, Feijen, van Gasteren: "Derivation of a termination
 -    detection algorithm for distributed computations", InfProc 
 -    Letters 16 (1983)]
 -
 - * LibRequest would be completely obsolete if Weak pointers would work
 -   as I'd expect them to, but unfortunately... [2000-09-30]
 - 
 ----------------------------------------------------------------------------
 - TODO:
 - * document this file properly
 - 
 ----------------------------------------------------------------------------
 -}

module TermCheckMPI
    ( initialize	-- :: Comm -> IO ()
    , isend		-- :: Serialize a => Rank -> Tag 
			--    -> Comm -> a -> IO ()
    , recv		-- :: Serialize a => Rank -> Tag
			--    -> Comm -> IO (Maybe (Status, a))
    , sendTermRequest	-- :: Comm -> IO ()
    )
where

import qualified HaskellMPI as HMPI
import qualified MPI
import MPI_SerializeDefault -- FIXME

import IOExts
import Monad
    ( when
    , liftM )
import Maybe
    ( mapMaybe
    , catMaybes
    , fromJust
    , isJust
    , isNothing )
import Weak 
    ( addFinalizer )

-----------------------------------------------------------------------------
data Mark = MarkInvalid | MarkValid | Master
    deriving (Eq)


type Request = LibRequest

data LibRequest = 
    LibRequest{ req   :: MPI.Request
              , doneR :: IO ()
              }


type GlobalVar a = IORef a

-----------------------------------------------------------------------------
-- library constants
magicTerminationTag :: HMPI.Tag
magicTerminationTag = HMPI.Tag 99


magicTermCheckTag :: HMPI.Tag
magicTermCheckTag = HMPI.Tag 98


-----------------------------------------------------------------------------
-- library variables

{-# NOINLINE libComm #-}
libComm :: GlobalVar HMPI.Comm
libComm = unsafePerformIO $ newIORef err
  where err = error "TermCheckMPI.libComm: not initialized"


{-# NOINLINE mark #-}
mark :: GlobalVar Mark
mark = unsafePerformIO $ newIORef MarkInvalid


{-# NOINLINE openRequests #-}
openRequests :: GlobalVar [Request]
openRequests = unsafePerformIO $ newIORef []


-----------------------------------------------------------------------------
initialize :: HMPI.Comm -> IO ()
initialize comm = do
    comm' <- HMPI.commDup comm
    libComm <-= comm'

    rank <- HMPI.commGetRank comm'
    mark <-= if rank == HMPI.Rank 0
               then Master
               else MarkInvalid


isend :: Serialize a => HMPI.Rank -> HMPI.Tag -> HMPI.Comm -> a -> IO ()
isend rank tag comm what = do
    breq <- HMPI.isend rank tag comm what

    let r = HMPI.toReq breq
    openRequests <-$ (:) (LibRequest r (wait__ breq))

  where
    wait__ br = HMPI.wait_ br >> return ()

     
recv :: Serialize a 
     => HMPI.Rank -> HMPI.Tag -> HMPI.Comm -> IO (Maybe (HMPI.Status, a))
-- receive specified message, or accept termination messages if
-- none is available.
recv rank tag comm = do
    comm' <- getLibComm
    self  <- liftM fromEnum (HMPI.commGetRank comm')

    handleOpenRequests

    -- irecvMax not possible here, since size of real msg unknown
    moreWork <- HMPI.iprobe rank tag comm

    -- termination attempt, if no real message is available
    -- this saves some superfluous termination messages
    when (isNothing moreWork)
         (tryTermination self) -- back door to escape deadlock
    
    -- wait until either library specific message or 
    -- "more work" message is available
#if 1
    ~[moreWork,termMsg] <- while (all isNothing) [moreWork,Nothing] $! do
        moreWork <- HMPI.iprobe rank tag comm
        termMsg  <- unsafeInterleaveIO 
                      (HMPI.iprobe HMPI.AnySource HMPI.AnyTag comm')
        return [moreWork, termMsg]
#else
    [moreWork,termMsg] <- repeatUntil (any isJust) $! do
        moreWork <- HMPI.iprobe rank tag comm
        termMsg  <- HMPI.iprobe HMPI.AnySource HMPI.AnyTag comm'
        return [moreWork, termMsg]
#endif

    case moreWork of
      Just stat -> do
        invalidateMark -- we're busy again
        let sFrom = HMPI.statusGetSource stat
            sTag  = HMPI.statusGetTag    stat
        liftM Just (HMPI.recv sFrom sTag comm)
     
      Nothing -> do
        let stat  = fromJust termMsg
            sFrom = HMPI.statusGetSource stat
            sTag  = HMPI.statusGetTag    stat
        m <- readIORef mark        
        (stat, msg) <- HMPI.recv sFrom sTag comm'
        handleTermCheck self comm' m stat msg

  where
    handleTermCheck self comm' markStatus status msg
      | magicTerminationTag == t = do
        -- termination NOW!
        return Nothing
#if 0
      | magicTermCheckTag   /= t = do
        -- never happens...
        error $ "TermCheckMPI.handleTermCheck: UNKNOWN TAG " ++ show t
#endif
      | markStatus /= MarkValid = do
        -- catch message and set self as new master
        mark <-= Master
        recv rank tag comm

      -- FROM HERE: markStatus == MarkValid!
      | msg == self = do
        -- termination NOW! But this process initiated it, so it
	-- has to tell the others...
        sendTermRequest comm'
        return Nothing

      | otherwise {- msg /= self -} = do
        -- continue this cycle
        forwardMessage msg
        recv rank tag comm

      where
      t = HMPI.statusGetTag status


    -- start a new termination cycle, but only if Master flag set
    tryTermination self = do
        isMaster <- liftM (Master ==) (readIORef mark)
        when isMaster 
             (forwardMessage self)


    -- set mark to `invalid' unless we're the master
    invalidateMark = do
        mark <-$ \ m ->
            if m == MarkValid 
              then MarkInvalid
              else m


    -- send termination request to next node (counter-clockwise, for
    -- cheaper comparisons) and drop master rights
    forwardMessage msg = do
        comm' <- getLibComm
        rank  <- HMPI.commGetRank comm'
        size  <- HMPI.commGetSize comm'
        
        let
          recipient | rank == HMPI.Rank 0 = HMPI.Rank (size-1)
                    | otherwise           = pred rank

        isend recipient magicTermCheckTag comm' msg
        mark <-= MarkValid

        return ()


    handleOpenRequests = do
        rs  <- readIORef openRequests
        -- FIXME: this is not portable, it relies on the MPI layer rather
        --        than on the HaskellMPI layer
        --        workaround: toplevel polymorphic(!) IORef, but this is 
        --                    ***extremely*** dangerous!
        rs' <- mapM MPI.test (map req rs)

        unfinished <- liftM catMaybes (mapM sync (zip rs' rs))
        openRequests <-= unfinished

        where
          sync (Nothing, y) = return (Just y)
          sync (Just _, y)  = doneR y >> return Nothing


sendTermRequest :: HMPI.Comm -> IO ()
-- send termination messages to all nodes except your own
sendTermRequest comm = do
    size <- HMPI.commGetSize comm
    self <- HMPI.commGetRank comm
    let ranks = [HMPI.Rank 0 .. HMPI.Rank (size-1)]
    mapM_ notify [r | r <- ranks, r /= self]

  where
    notify r = HMPI.send r magicTerminationTag comm ({-don't care-}42::Int)


-----------------------------------------------------------------------------
-- utils
getLibComm :: IO HMPI.Comm
getLibComm = readIORef libComm


while cond val p = loop' val
   where
   loop' v 
     | cond v = do
       v' <- p
       loop' v'

     | otherwise = do
       return v


repeatUntil cond p = loop'
    where
    loop' = do
        v <- p
        if cond v then return v
                  else loop'


infix 3 <-=, <-$
(<-=) = writeIORef
(<-$) = modifyIORef


-----------------------------------------------------------------------------
